# fis3-hook-amd

fis3 已经默认不自带模块化开发支持，那么如果需要采用 amd 规范作为模块化开发，请使用此插件。

## 安装

全局安装或者本地安装都可以。

```
npm install -g fis3-hook-amd
```

或者

```
npm install fis3-hook-amd
```

## 用法

在 fis-conf.js 中加入以下代码。


```js
fis.hook('amd', {
  // 配置项
});
```

## 说明

请配合 `require.js`、`esl.js` 或者其他满足 amd 规范的前端加载器一起使用。

注意：虽然是 amd 插件，但是同样支持 commonjs 规范，对目标文件设置 `isMod` 属性，此插件会自动包裹成 `amd` 规范的。


```js
fis.match('/modules/**.js', {
  isMod: true
})
``` 

另外需要注意的是 amd 插件需要，需要把目标文件解析成 `ast`，然后再基于 `ast` 来处理，对于大的 js 文件来说，确实很费时间。那么如果想跳过某些文件的分析，请添加 `skipDepsAnalysis:true`, 可以大大减少编译时间。

```js
fis.match('some big file', {
  skipDepsAnalysis: true
})
```

## 自定义模块ID

插件默认使用资源的绝对路径作为模块ID，如果希望更改模块ID，需要对目标文件设置 `moduleId` 属性

```js
fis.match('/modules/(**).js', {
  isMod: true,
  moduleId: '$1'
})
``` 

通过上述配置，我们可以将 `/modules/A.js` 的模块ID由 `modules/A` 改变为 `A`

**另外**

fis 的 amd 方案，是把对依赖的分析过程从运行期改成了编译期，所以请尽量不要设置 `require.config({options...})`, 因为一旦设置了 `baseUrl` 、 `paths` 或者 `packages` 什么的，会让 `fis` 静态编译时分析变得很困难，甚至分析不到。

但是，你可以在编译期做同样的配置。

// in fis-conf.js

```
fis.hook('amd'{
  baseUrl: './modules',
  paths: {
    $: 'jquery/jquery-1.11.2.js'
  }
})
```

具体请查看[配置项说明](#配置项)。

## 配置项

* `globalAsyncAsSync` 是否将全局下面的异步用法，当同步处理。作用是，本来要运行时加载的，现在变成页面里面直接引用了。

  此开关为全局选项，如果只想控制部分 `require` 的用法，可以通过 js 注释来控制。
  
  当全局异步用法当同步用法关闭时，只有加了 `fis sync` 注释的异步用法才会当同步处理。
  
  ```js
  // fis sync
  require(['./module/a'])
  ```
  
  当全局异步用法当同步用法开启时，只有加了 `fis async` 注释的异步用法才保留异步作用。
  
  ```js
   // fis async
  require(['./module/a'])
  ```
   
* `baseUrl` 默认为 `.` 即项目根目录。用来配置模块查找根目录。
* `paths` 用来设置别名，路径基于 `baseUrl` 设置。
  
  ```js
  fis.hook('amd', {
    paths: {
      $: '/modules/jquery/jquery-1.11.2.js'
    }
  });
  ```
* `packages` 用来配置包信息，方便项目中引用。
  
  ```js
  fis.hook('amd', {
    packages: [
      {
        name: 'foo',
        location: './modules/foo',
        main: 'index.js'
      }
    ]
  });
  ```

  * 当 `require('foo')` 的时候等价于 `require('/modules/foo/index.js')`.
  * 当 `require('foo/a.js')` 的时候，等价于 `require('/modules/foo/a.js')`.
* `shim` 可以达到不改目标文件，指定其依赖和暴露内容的效果。**注意只对不满足amd的js有效**
  
  ```js
  fis.hook('amd', {
      shim: {
          'comp/2-0/2-0.js': {
              deps: ['jquery'],
              exports: 'myFunc'
          }
      }
  });
  ```
  
  * `key` 为目标文件
  * `value`
    * `deps` [可选] 依赖的 `module` 列表。
    * `exports` [可选] 暴露的对象名称。
    * `init` [可选] 暴露的可以通过自定的方法来控制。
    
      ```js
      fis.hook('amd', {
          shim: {
              'comp/2-0/2-0.js': {
                  deps: ['jquery'],
                  init: 'function($) {return $.extend({a: 1}, {b: 2})}'
              }
          }
      });
      ```
* `forwardDeclaration` 默认为 `true`, 用来设置是否开启依赖前置，根据前端加载器来定。
* `skipBuiltinModules` 默认为 `false`, 只有在 `forwardDeclaration` 启动的时候才有效，用来设置前置依赖列表中是否跳过内置模块如： `require`, `module`, `exports`。
* `extList` 默认为 `['.js', '.coffee', '.jsx', '.es6']`，当引用模块时没有指定后缀，该插件会尝试这些后缀。
* `tab` 默认为 `2`, 用来设置包裹时，内容缩进的空格数。
* `ignoreDependencies` 默认为空，当分析到某个文件的时候，此插件会把当前文件标记依赖目标文件。如果你希望部分文件不这么做，那么请设置此插件。

  ```js
  fis.hook('amd', {
      ignoreDependencies: [
        'angular2/angular2'
      ]
  });
  ```
